#----- import token
from enum import Enum
import socket
import time
import sys
import os
import serial ### pyserial のインストールが必要

class Tk(Enum):
	(RESWD, NUMLIT, VAR, LITERAL, COMM, END) = range(6)

restbl = [
	"print",
	"clear",
	"goto",
	"if",
	"for",
	"to",
	"step",
	"next",
	"then",
	"else",
	"endif",
	"input",
	"rem",
	"dim",
	"gosub",
	"return",
	"end",
]

class Token:
	'''トークン'''
	def __init__(self, type, image, pos):
		self.type = type
		self.image = image
		self.pos = pos
	
	def isRes(self, st):
		return self.type == Tk.RESWD and self.image == st
		
class Que:
	def __init__(self, st):
		self.buff = st + "\n"
		self.pos = -1
		
	def pop(self):
		self.pos += 1
		return self.buff[self.pos]
		
	def currPos(self):
		return self.pos

def print_msg(msg, pos):
	print("-" * (pos+1) + "↑")
	print(msg)

def tkn_variable():
	'''
	変数トークンの作成
	{a-z|A-Z|_}[a-z|A-Z|_|0-9]…
	'''
	global inp, ch

	startPos = inp.currPos()
	img = ""
	
	while "a" <= ch <= "z" or "A" <= ch <= "Z" or ch == "_" or \
		  "0" <= ch <= "9":
		img += ch
		ch = inp.pop()
		
	if ch == "$":
		img += ch
		ch = inp.pop()
	return Token(Tk.VAR, img, startPos)


def tkn_numlit():

	'''
	数字定数トークンの作成
		[+|-] { {0-9}… | [0-9]… .{0-9}… }
	'''
	global inp, ch
	
	decimal_point = False
	integer = 0
	decimal = 0
	image = ""
	state = 0
	
	startPos = inp.currPos()
	
	while "0" <= ch <= "9" or ch in ".+-":
		if state == 0:
			if ch in "+-":
				state = 1
				image += ch
				ch = inp.pop()
				continue
			state = 1
			continue
		if state == 1:
			if "0" <= ch <= "9":
				integer += 1
				image += ch
				state = 1
				ch = inp.pop()
				continue
			if ch == ".":
				decimal_point = True
				image += ch
				state = 2
				ch = inp.pop()
				continue
			break
		if state == 2:
			if "0" <= ch <= "9":
				decimal += 1
				image += ch
				state = 2
				ch = inp.pop()
				continue
			break

	if image == "+" or image == "-":
		return Token(Tk.RESWD, image, startPos)

	if decimal_point and decimal == 0:
		print_msg("L001: 数字として正しくない", startPos)
		return None
	if not decimal_point and integer == 0:
		print_msg("L002: 数字として正しくない", startPos)
		return None

	return Token(Tk.NUMLIT, image, startPos)
	
def tkn_literal():
	'''
	文字列トークンの作成
	'''
	global inp, ch
	
	image = ch
	state = 0
	
	startPos = inp.currPos()
	ch = inp.pop()

	while ch != '\n':
		if ch == '"':
			image += ch
			ch = inp.pop()
			return Token(Tk.LITERAL, image, startPos)
		image += ch
		ch = inp.pop()
		
	print_msg("L004: 文字列の終わりがない", startPos)
	return None
		
def make_token(st):

	global inp, ch
	
	tokens = []

	inp = Que(st)
	
	ch = inp.pop()
	
	while ch != "\n":
	
		if ch in "*/()=,;":
			tokens.append(Token(Tk.RESWD, ch, inp.currPos()))
			ch = inp.pop()
			continue
		if ch in "+-":
			if len(tokens) > 0:
				tkn = tokens[-1]
				if tkn.type == Tk.NUMLIT or tkn.type == Tk.VAR or tkn.isRes(")"):
				   	tokens.append(Token(Tk.RESWD, ch, inp.currPos()))
				   	ch = inp.pop()
				   	continue
			tkn = tkn_numlit()
			if tkn != None:
				tokens.append(tkn)
			else:
				return None
			continue
		if ch == "<":
			ch = inp.pop()
			if ch == "=" or ch == ">":
				tokens.append(Token(Tk.RESWD, "<"+ch, inp.currPos()-1))
				ch = inp.pop()
				continue
			tokens.append(Token(Tk.RESWD, "<", inp.currPos()-1))
			continue
		if ch == ">":
			ch = inp.pop()
			if ch == "=":
				tokens.append(Token(Tk.RESWD, ">=", inp.currPos()-1))
				ch = inp.pop()
				continue
			tokens.append(Token(Tk.RESWD, ">", inp.currPos()-1))
			continue
		if "0" <= ch <= "9" or ch == ".":
			tkn = tkn_numlit()
			if tkn != None:
				tokens.append(tkn)
			else:
				return None
			continue
		if ch == '"':
			tkn = tkn_literal()
			if tkn != None:
				tokens.append(tkn)
			else:
				return None
			continue
		if "a" <= ch <= "z" or "A" <= ch <= "Z" or ch == "_":
			tkn = tkn_variable()
			if tkn != None:
				if tkn.image in restbl:
					tkn.type = Tk.RESWD
				tokens.append(tkn)
				if tkn.isRes("rem"):
					ch = inp.pop()
					comm_img = ""
					while ch != "\n":
						comm_img += ch
						ch = inp.pop()
					tokens.append(Token(Tk.COMM, comm_img, -1))
			else:
				return None
			continue
		if ch == " ":
			ch = inp.pop()
			continue
			
		print_msg("L003: 許されない文字がある:'"+ch+"'", inp.currPos())
		return None
	
	tokens.append(Token(Tk.END,"", -1))
	return tokens	
#----- import program
class Program:
	'''
	プログラムを保存するクラス
		lines  : 行番号を管理するリスト  [行番号…]
		program: プログラムを管理する辞書
		            key = 行番号, value = [トークン列…]
		counter: 実行するプログラム行番号(lines上のインデックス値)
		
		例）  10 a = 1
		      20 b = 1
		      30 print a + b
		   の場合は、
		      program = { 10: [a=1のトークン列],
		                  20: [b=1のトークン列],
		                  30: [print a+bのトークン列] }
		      lines = [ 10, 20, 30 ]
	'''
	def __init__(self):
		self.clear()
		return
		
	def clear(self):
		self.program = {}
		self.lines = []
		self.counter = 0
		return
		
	def put(self, line, tkns):
		self.program[line] = tkns
		if line in self.lines:
			return
		self.lines.append(line)
		self.lines.sort()
		return
		
	def set_first(self):
		if len(self.lines) == 0:
			return -1
		self.counter = 0
		return 0
		
	def set_next(self):
		self.counter += 1
		if self.counter >= len(self.lines):
			return -1
		return 0
		
	def set_line(self, line):
		if line in self.lines:
			ind = self.lines.index(line)
			self.counter = ind
			return 0
		return -1
		
	def get_line(self):
		return (self.lines[self.counter], 
			self.program[self.lines[self.counter]][:])
			
	def line_image(self, nest=0):
		#st = ""
		num, tkns = self.get_line()
		tkn = tkns.pop(0)
		if tkn.type == Tk.RESWD:
			st = tkn.image + " "
		else:
			st = tkn.image
		for tkn in tkns:
			if (len(tkn.image)==1 and tkn.image in "+-*/=><") or\
				tkn.isRes(">=") or tkn.isRes("<=") or tkn.isRes("<>") or\
				tkn.isRes("to") or tkn.isRes("then") or\
				tkn.isRes("goto") or tkn.isRes("step"): #★★★
				st += " " + tkn.image + " "
			elif ("a" <= st[-1] <="z" or "A" <= st[-1] <= "Z" or \
				"0" <= st[-1] <= "9" or st[-1] == "_") and\
				tkn.type in [Tk.VAR, Tk.NUMLIT, Tk.LITERAL]:
					st += " " + tkn.image
			else:
				st += tkn.image
		return "{0:5d} ".format(num) + nest * "  " + st
		
	def del_line(self, line):
		if self.set_line(line) == -1:
			return
		del self.program[self.lines[self.counter]]
		del self.lines[self.counter]
		return
#-----
	
class Stack:
	'''スタック'''
	
	def __init__(self):
		self.buff = []
		
	def push(self, st):
		self.buff.insert(0, st)
		
	def pop(self):
		if len(self.buff) == 0:
			print("Stack Error")
			return None
		return self.buff.pop(0)
	
	def size(self):
		return len(self.buff)
		
	def print(self):
		print(self.buff)

def expression_lit(tokens):
	'''
	文字列の演算
	'''
	global print_stk, assign
	
	state = 0
	tkn = tokens.pop(0)
	ret_image = ""
	
	while tkn.type != Tk.END:
		if state == 0:
			if tkn.type == Tk.VAR:
				if tkn.image[-1] != "$":
					print_err("E007: 文字列変数でない: "+tkn.image)
					return None
				if tkn.image == "mid$":
					tkn = tokens.pop(0)
					if not tkn.isRes("("):
						print_err("E014:括弧「(」でなければならない: "+tkn.image)
						return None
					ret, tkns = copy_until(tokens,[","])
					if ret == -1:
						print_err("E015:第一引数の終わりのカンマが見つからない: mid$")
						return None
					mid_string = expression_lit(tkns)
					if mid_string == None:
						return None
					mid_string = mid_string[1:-1]
					tkn = tokens.pop(0) #カンマの読み飛ばし
					ret, tkns = copy_until(tokens,[",",")"])
					if ret == -1:
						print_err("E016:第二引数の終わりのカンマ「,」または右括弧「)」が見つからない: mid$")
						return None
					mid_start = expression(tkns)
					if mid_start == None:
						return None
					mid_start = int(mid_start)
					if mid_start <= 0:
						print_err("E017:開始位置は1以上の整数でなければならない: mid$")
						return None
					tkn = tokens.pop(0)
					if tkn.isRes(","):
						# 長さ指定あり
						ret, tkns = copy_until(tokens, [")"])
						if ret == -1:
							print_err("E018:右括弧が見つからない: mid$")
							return None
						mid_length = expression(tkns)
						if mid_length == None:
							return None
						mid_length = int(mid_length)
						if mid_length < 0:
							print_err("E019:長さ指定は0以上の整数でなければならない: mid$")
							return None
						tkn = tokens.pop(0) #「)」の読み飛ばし
					else:
						mid_length = len(mid_string) - mid_start + 1
					if len(mid_string) >= mid_start:
						ret_image += mid_string[mid_start-1:mid_start-1+mid_length]
					tkn = tokens.pop(0)
					state = 1
					continue
				if tkn.image not in assign.keys():
					print_err("E008: 未定義の文字列変数である: "+tkn.image)
					return None
				val = get_var_value(tkn, tokens)
				if val == None:
					return None
				ret_image += val[1:-1]
				tkn = tokens.pop(0)
				state = 1
				continue
			if tkn.type == Tk.LITERAL:
				ret_image += tkn.image[1:-1]
				tkn = tokens.pop(0)
				state = 1
				continue
			print_err("E009: 文字列定数/変数でなければならない: "+tkn.image)
			return None
		if state == 1:
			if tkn.isRes("+"):
				tkn = tokens.pop(0)
				state = 0
				continue
			print_err("E010: 演算子(+)でない: "+tkn.image)
			return None
			
	if state == 1:
		return '"' + ret_image + '"'
		
	print_err("E011: 文字列の演算が正しくない")
	return None

def get_var_value(var, tokens):

	global assign, var_dim
	
	tkn = tokens.pop(0)
	if not tkn.isRes("("):
		if var.image in var_dim.keys():
			print_err("B001:配列指定でなければならない: "+var.image)
			return None
		tokens.insert(0, tkn)
		return assign[var.image]
		
	#配列の参照
	#  参照：   変数 '(' 添字 [, 添字] ')'
	#  添字：     { 数字変数 | 数字定数 }
	if not var.image in var_dim.keys():
		print_err("B002:配列として定義されていない: "+var.image)
		return None
	tkn = tokens.pop(0)
	dim = []
	state = 0
	while True:
		if tkn.type == Tk.END:
			print_err("B003:配列の添字指定が正しくない: "+var.image)
			return None
		if state == 0:
			tokens.insert(0, tkn)
			ret, tkns = copy_until(tokens,[",",")"])
			if ret == -1:
				print_err("B011:「)」または「,」でなければならない: "+tkn.image)
				return None
			ans = expression(tkns)
			if ans == None:
				return None
			dim.append(int(ans))
			if len(dim) > 2:
				print_err("B006:3次元以上は指定できない: "+var.image)
				return None
			state = 1
			tkn = tokens.pop(0)
			continue
		if state == 1:
			if tkn.isRes(","):
				tkn = tokens.pop(0)
				state = 0
				continue
			if tkn.isRes(")"):
				if len(dim) != len(var_dim[var.image]):
					print_err("B009:定義されている次元数("+str(len(var_dim[var.image]))+")が指定されていない: "+var.image)
					return None
				for i, j in zip(var_dim[var.image], dim):
					if i < j or j <= 0:
						print_err("B010:添字の値が1から最大値("+str(int(i))+")の範囲ではない: "+str(int(j)))
						return None
				#tkn = tokens.pop(0)
				break
			print_err("B011:「)」または「,」でなければならない: "+tkn.image)
			return None
	
	if len(dim) == 1:
		return assign[var.image][dim[0]-1]
	else:
		return assign[var.image][dim[0]-1][dim[1]-1]
			
def get_int_value(tokens):

	tkn = tokens.pop(0)
	if not tkn.isRes("("):
		print_err("M001:左括弧「(」でなければならない: "+tkn.image)
		return None
	ret, tkns = copy_until(tokens, [")"])
	if ret == -1:
		print_err("M002:右括弧「)」が見つからない: "+tkn.image)
		return None
	val = expression(tkns)
	if val == None:
		return None
	tmp = int(val)
	tkn = tokens.pop(0)
	if tmp > val:
		return tmp - 1
	else:	
		return tmp
		
def expression(tokens):

	'''
	数字1または変数1 { {+|-|*|/} 数字2または変数2 | = }…
	0                  1         2               1  1
	'''
	global print_stk, assign
	
	stk = Stack()
	paren = 0
	state = 0
	
	tkn = tokens.pop(0)
	
	while tkn.type != Tk.END:
		if print_stk:
			stk.print()
		if state == 0:
			#「(」の入力
			if tkn.isRes("("):
				paren += 1
				stk.push("(")
				state = 0
				tkn = tokens.pop(0)
				continue
			# 単項演算子 +, - の場合
			if tkn.isRes("+"):
				state = 0
				tkn = tokens.pop(0)
				continue
			if tkn.isRes("-"):
				tokens.insert(0, Token(Tk.RESWD, "*", -1))
				tokens.insert(0, Token(Tk.NUMLIT, "-1", -1))
				state = 0
				tkn = tokens.pop(0)
				continue
			# 数字1または変数1の入力
			if tkn.type == Tk.VAR:
				if tkn.image[-1] == "$":
					print_err("E012:文字列変数を指定することはできない: "+tkn.image)
					return None
				if tkn.image == "int":
					val = get_int_value(tokens)
					if val == None:
						return None
					stk.push(val)
					state = 1
					tkn = tokens.pop(0)
					continue
				elif tkn.image in assign.keys():
					val = get_var_value(tkn, tokens)
					if val == None:
						return None
					stk.push(float(val))
					state = 1
					tkn = tokens.pop(0)
					continue
				else:
					print_err("E013:未定義の変数である: "+tkn.image)
					return None
			if tkn.type != Tk.NUMLIT:
				print_err("E001:数字変数または数字定数でなければならない: "+tkn.image)
				return None
			stk.push(float(tkn.image))
			state = 1
			tkn = tokens.pop(0)
			continue
		if state == 1:
			#「)」の入力
			if tkn.isRes(")"):
				if paren == 0:
					print_err("E002:括弧の対応がとれていない: "+tkn.image)
					return None
				paren -= 1
				b = stk.pop()
				op = stk.pop()
				if op in "+-":
					a = stk.pop()
					if op == "+":
						a += b
					else:
						a -= b
					b = a
					stk.pop() #「(」の読み飛ばし
				else: #「(」のはず
					pass
				if stk.size() == 0:
					stk.push(b)
					state = 1
					tkn = tokens.pop(0)
					continue
				op = stk.pop()
				if op in "+-*/":
					a = stk.pop()
					if op == "+":
						a += b
					if op == "-":
						a -= b
					if op == "*":
						a *= b
					if op == "/":
						a /= b
					stk.push(a)
					state = 1
					tkn = tokens.pop(0)
					continue
				#opが「(」の場合
				stk.push(op)
				stk.push(b)
				state = 1
				tkn = tokens.pop(0)
				continue
			if tkn.type != Tk.RESWD or tkn.image not in "+-*/":
				print_err("E003:演算子(+,-,*,/)ではない: "+tkn.image)
				return None
			op = tkn.image
			if op in "*/":
				stk.push(op)
			else: #+, -
				if stk.size() >= 3:
					b = stk.pop()
					op1 = stk.pop()
					if op1 in "+-":
						a = stk.pop()
						if op1 == "+":
							a += b
						else:
							a -= b
						stk.push(a)
						stk.push(op)
					else: #「(」のはず
						stk.push(op1)
						stk.push(b)
						stk.push(op)
				else:
					stk.push(op)
			state = 2
			tkn = tokens.pop(0)
			continue
		if state == 2:
			#「(」の入力
			if tkn.isRes("("):
				paren += 1
				stk.push("(")
				state = 0
				tkn = tokens.pop(0)
				continue
			# 単項演算子 +, - の場合
			if tkn.isRes("+"):
				state = 2
				tkn = tokens.pop(0)
				continue
			if tkn.isRes("-"):
				tokens.insert(0, Token(Tk.RESWD, "*", -1))
				tokens.insert(0, Token(Tk.NUMLIT, "-1", -1))
				state = 2
				tkn = tokens.pop(0)
				continue
			#数字2/変数2の入力
			if tkn.type == Tk.VAR:
				if tkn.image[-1] == "$":
					print_err("E012:文字列変数を指定することはできない: "+tkn.image)
					return None
				if tkn.image == "int":
					val = get_int_value(tokens)
					if val == None:
						return None
					b = val
				elif tkn.image in assign.keys():
					val = get_var_value(tkn, tokens)
					if val == None:
						return None
					b = float(val)
				else:
					print_err("E013:未定義の変数である: "+tkn.image)
					return None
			elif tkn.type == Tk.NUMLIT:
				b = float(tkn.image)
			else:
				print_err("E001:数字変数または数字定数でなければならない: "+tkn.image)
				return None
			op = stk.pop()
			a = stk.pop()
			if op in "*/":
				if op == "*":
					a *= b
				else:
					a /= b
				stk.push(a)
			else:
				stk.push(a)
				stk.push(op)
				stk.push(b)
			state = 1
			tkn = tokens.pop(0)
			continue

	if state != 1:
		print_err("E005:算術式が誤っている")
		return None
		
	if paren > 0:
		print_err("E006:括弧の対応がとれていない")
		return None
		
	if stk.size() == 1:
		a = stk.pop()
	else:
		b = stk.pop()
		op = stk.pop()
		a = stk.pop()
		if op == "+":
			a += b
		else:
			a -= b
	return a
	
def copy_until(tokens, imglist):

	ret = []
	paren = 0
	tkn = tokens.pop(0)
	while tkn.type != Tk.END:
		if tkn.type == Tk.RESWD and tkn.image in imglist and paren == 0:
			tokens.insert(0, tkn)
			ret.append(Token(Tk.END, "", -1))
			return 0, ret
		ret.append(tkn)
		if tkn.isRes("("):
			paren += 1
		if tkn.isRes(")"):
			if paren > 0:
				paren -= 1
		tkn = tokens.pop(0)
		
	ret.append(Token(Tk.END, "", -1))	
	return -1, ret
	
def search_next(currline):

	global prg
	
	nest = 0
	prg.set_line(currline)
	
	while True:
		if prg.set_next() == -1:
			return None
		line, tokens = prg.get_line()
		tkn = tokens.pop(0)
		if tkn.isRes("for"):
			nest += 1
			continue
		if tkn.isRes("next"):
			if nest == 0:
				return line
			nest -= 1
			continue
	
def print_err(msg):

	global prg, running_flg
	
	if running_flg:
		print(prg.line_image())
	print(msg)
	sendMsg(msg)
	
def search_else_endif(currline):

	global prg
	
	nest = 0
	else_line = None
	endif_line = None
	prg.set_line(currline)
	
	while True:
		if prg.set_next() == -1:
			return None, None
		line, tokens = prg.get_line()
		tkn = tokens.pop(0)
		if tkn.isRes("if"):
			if tokens[-2].isRes("then"): #★★★
				nest += 1
				continue
		if tkn.isRes("endif"):
			if nest == 0:
				return else_line, line
			nest -= 1
			continue
		if tkn.isRes("else"):
			if nest == 0:
				if else_line == None:
					else_line = line
				else: # 既にelseがある
					return None, None
	
def cond_exp(tokens, stopper):

	tkn = tokens[0]
	if tkn.type == Tk.VAR and tkn.image[-1] == "$" or \
	   tkn.type == Tk.LITERAL:
		exp_proc = expression_lit
	else:
		exp_proc = expression
	ret, lside = copy_until(tokens,["=", "<>", ">", "<", ">=", "<="])
	if ret == -1:
		print_err("C001:条件式が正しくない(左辺)")
		return -1, None
	lval = exp_proc(lside)
	if lval == None:
		return -1, None
	tkn = tokens.pop(0)
	cmp = tkn.image
	ret, rside = copy_until(tokens, stopper)
	if ret == -1:
		print_err("C002:条件式が正しくない(右辺)")
		return -1, None
	rval = exp_proc(rside)
	if rval == None:
		return -1, None
	if cmp == "=": 	cmp = "=="
	if cmp == "<>":	cmp = "!="
	form = "{0}{1}{2}".format(lval, cmp, rval)
	condition = eval(form)
	return 0, condition
			
def statement(tokens):

	global assign, prg, trace_flg, running_flg, var_dim
	inputMsg=""

	if trace_flg and running_flg:
		print("exec: "+prg.line_image())
	
	tkn = tokens.pop(0)
	
	if tkn.isRes("rem"):
		return 0
		
	if tkn.isRes("end"):
		return -1
		
	if tkn.isRes("gosub"):
		tkn = tokens.pop(0)
		if tkn.type != Tk.NUMLIT:
			print_err("G001:行番号でなければならない: "+tkn.image)
			return -1
		sv_line, _ = prg.get_line()
		if "." in tkn.image:
			print_err("G002:行番号は正の整数でなければならない: "+tkn.image)
			return -1
		if prg.set_line(int(tkn.image)) == -1:
			print_err("G003:この行番号は存在しない: "+tkn.image)
			return -1
		_, tokens = prg.get_line()
		while True:
			tkn = tokens[0]
			if tkn.isRes("return"):
				if prg.set_line(sv_line) == -1:
					print_err("インタプリタエラー：gosub")
					return -1
				return 0
			ret = statement(tokens)
			if ret == 0:
				if prg.set_next() == -1:
					prg.set_line(sv_line)
					print_err("G004: return文が見つからない")
					return -1
				_, tokens = prg.get_line()
				continue
			if ret == -1:
				return -1
			if prg.set_line(ret) == -1:
				print_err("G003:この行番号は存在しない: "+tkn.image)
				return -1
			_, tokens = prg.get_line()
			continue
		return -1
		
	if tkn.isRes("dim"):
		tkn = tokens.pop(0)
		if tkn.type != Tk.VAR:
			print_err("D001:変数の指定でなければならない: "+tkn.image)
			return -1
		var_img = tkn.image
		tkn = tokens.pop(0)
		if not tkn.isRes("("):
			print_err("D002:配列要素数の指定が正しくない: "+var_img)
			return -1
		tkn = tokens.pop(0)
		dim = []  # dim a(n,m) と定義されていたら dim = [n, m] となる。
		state = 0
		while True:
			if tkn.type == Tk.END:
				print_err("D002:配列要素数の指定が正しくない: "+var_img)
				return -1
			if state == 0:
				if tkn.type != Tk.NUMLIT:
					print_err("D003:配列要素数は正の整数でなければならない: "+tkn.image)
					return -1
				if "." in tkn.image:
					print_err("D003:配列要素数は正の整数でなければならない: "+tkn.image)
					return -1
				if int(tkn.image) <= 0:
					print_err("D003:配列要素数は正の整数でなければならない: "+tkn.image)
					return -1
				dim.append(int(tkn.image))
				tkn = tokens.pop(0)
				state = 1
				continue
			if state == 1:
				if tkn.isRes(","):
					tkn = tokens.pop(0)
					state = 0
					continue
				if tkn.isRes(")"):
					if len(dim) >= 3:
						print_err("D004:３次元以上の配列は定義できない: "+var_img)
						return -1
					var_dim[var_img] = dim
					assign[var_img] = {}
					if len(dim) == 1:
						for i in range(dim[0]):
							if var_img[-1] == "$":
								assign[var_img][i] = ""
							else:
								assign[var_img][i] = 0
					else:
						#2次元
						for i in range(dim[0]):
							assign[var_img][i] = {}
							for j in range(dim[1]):
								if var_img[-1] == "$":
									assign[var_img][i][j] = ""
								else:
									assign[var_img][i][j] = 0
					return 0
				print_err("D005:「,」または「)」でなければならない: "+tkn.image)
				return -1
		return -1
					
	if tkn.isRes("input"):
		tkn = tokens.pop(0)
		if tkn.type == Tk.LITERAL:
			prompt = tkn.image[1:-1]
			tkn = tokens.pop(0)
			state = 1
		else:
			prompt = ""
			state = 0
		vars = []
		while tkn.type != Tk.END:
			if state == 0:
				if tkn.type == Tk.VAR:
					vars.append(tkn.image)
					tkn = tokens.pop(0)
					state = 1
					continue
				print_err("I001:変数が指定されていない: "+tkn.image)
				return -1
			if state == 1:
				if tkn.isRes(","):
					tkn = tokens.pop(0)
					state = 0
					continue
				print_err("I002:カンマが指定されていない: "+tkn.image)
				return -1
		if len(vars) == 0:
			print_err("I003:変数が指定されていない")
			return -1
		st = input(prompt+"? ")
		assignVars=""
		vals = st.split()
		if len(vars) != len(vals):
			print_err("I004: 入力項目に過不足がある")
			line, _ = prg.get_line()
			return line #再入力
		for var, val in zip(vars, vals):
			if var[-1] == "$": #文字列変数
				assign[var] = '"'+str(val)+'"'
				assignVars+=str(val)+"\n"
			elif val.isnumeric():
				assign[var] = float(val)
				assignVars+=str(val)+"\n"
			else:
				print_err("I005:変数の型と入力の値が一致しない: "+var+" <==> "+str(val))
				line, _ = prg.get_line()
				return line #再入力
		sendMsg(prompt+"? "+assignVars)
		return 0
	
	if tkn.isRes("for"):
		tkn = tokens.pop(0)
		if tkn.type != Tk.VAR:
			print_err("F001:数字変数でなければならない: "+tkn.image)
			return -1
		if tkn.image[-1] == "$":
			print_err("F001:数字変数でなければならない: "+tkn.image)
			return -1
		var = tkn.image
		tkn = tokens.pop(0)
		if not tkn.isRes("="):
			print_err("F002:変数の次は「=」でなければならない: "+tkn.image)
			return -1
		ret, exps = copy_until(tokens,["to"])
		if ret == -1:
			print_err("F003:「to」指定がない")
			return -1
		init_val = expression(exps)
		if init_val == None:
			print_err("F004:初期値の算術式が正しくない")
			return -1
		tkn = tokens.pop(0) # toの読み飛ばし
		ret, exps = copy_until(tokens, ["step"])
		to_val = expression(exps)
		if to_val == None:
			print_err("F005:終了値の算術式が正しくない")
			return -1
		if ret == 0: # step指定がある
			tkn = tokens.pop(0) #step文の読み飛ばし
			step_val = expression(tokens)
			if step_val == None:
				print_err("F006:stepで指定している加算値が正しくない")
				return -1
		else:
			step_val = 1
		for_line , _ = prg.get_line()
		next_line = search_next(for_line)
		prg.set_line(for_line)
		if next_line == None:
			print_err("F007:対応するnext文が見つからない")
			return -1
		if prg.set_next() == -1:
			print_err("インタプリタエラー:for文")
			return -1
		assign[var] = init_val
		cont_line, tokens = prg.get_line()
		while True:
			tkn = tokens.pop(0)
			if tkn.isRes("next"):
				assign[var] += step_val
				if step_val > 0 and assign[var] > to_val:
					return 0
				if step_val < 0 and assign[var] < to_val:
					return 0
				if prg.set_line(cont_line) == -1:
					print_err("インタプリタエラー：for文(2)")
					return -1
				_, tokens = prg.get_line()
				continue
			tokens.insert(0, tkn)
			ret = statement(tokens)
			if ret == 0:
				if prg.set_next() == -1:
					print_err("インタプリタエラー：for文(3)")
					return -1
				_, tokens = prg.get_line()
				continue
			if ret == -1: #終了
				return -1
			if for_line < ret <= next_line:
				if prg.set_line(ret) == -1:
					print_err("F008:この行番号は存在しない: "+str(ret))
					return -1
				_, tokens = prg.get_line()
				continue
			return ret
	
	if tkn.isRes("if"):
		ret, condition = cond_exp(tokens, ["goto", "then"])
		if ret == -1:
			return -1
		tkn = tokens.pop(0)
		# goto/thenトークンで処理を変える。
		if tkn.isRes("goto"):
			if condition:
				tkn = tokens.pop(0)
				if tkn.type != Tk.NUMLIT:
					print_err("H001:行番号は正の整数でなければならない: "+tkn.image)
					return -1
				if "." in tkn.image:
					print_err("H001:行番号は正の整数でなければならない: "+tkn.image)
					return -1
				return int(tkn.image)
			return 0
		# thenの場合
		if_line, tokens = prg.get_line()
		else_line, endif_line = search_else_endif(if_line)
		prg.set_line(if_line)
		if endif_line == None:
			print_err("H002:対応するendif文がない、またはelse文が2つ以上ある")
			return -1
		if not condition:
			if else_line == None:
				prg.set_line(endif_line)
				return 0
			prg.set_line(else_line)
		if prg.set_next() == -1:
			print_err("インタプリタエラー:IF文")
			return -1
		while True:
			_, tokens = prg.get_line()
			tkn = tokens.pop(0)
			if tkn.isRes("else"):
				prg.set_line(endif_line)
				return 0
			if tkn.isRes("endif"):
				return 0
			tokens.insert(0, tkn)
			ret = statement(tokens)
			if ret == 0:
				if prg.set_next() == -1:
					print_err("インタプリタエラー:IF文(2)")
					return -1
				continue
			if ret == -1:
				return -1
			if if_line < ret <= endif_line:
				if prg.set_line(ret) == -1:
					print_err("H003:この行番号は存在しない: "+str(ret))
					return -1
				continue
			return ret
	
	if tkn.isRes("goto"):
		tkn = tokens.pop(0)
		if tkn.type != Tk.NUMLIT:
			print_err("J001:行番号は正の整数でなければならない: "+tkn.image)
			return -1
		if "." in tkn.image:
			print_err("J001:行番号は正の整数でなければならない: "+tkn.image)
			return -1
		return int(tkn.image)
	
	if tkn.isRes("clear"):
		assign = {}
		var_dim = {}
		return 0
	
	if tkn.isRes("print"):
		print_img = ""
		state = 0
		tkn = tokens.pop(0)
		while tkn.type != Tk.END:
			if state == 0:
				if tkn.type == Tk.VAR and tkn.image[-1] == "$" or \
				   tkn.type == Tk.LITERAL:
					tokens.insert(0, tkn)
					ret, exps = copy_until(tokens, [";",","])
					ans = expression_lit(exps)
					if ans == None:
						return -1
					ans = ans[1:-1]
				else:
					tokens.insert(0, tkn)
					ret, exps = copy_until(tokens,[";",","])
					ans = expression(exps)
					if ans == None:
						return -1
					ans = str(ans)
				print_img += ans
				if ret == -1:
					print(print_img)
					sendMsg(print_img)
					return 0
				tkn = tokens.pop(0)
				state = 1
				continue
			if state == 1:
				if tkn.isRes(";"):
					pass
				elif tkn.isRes(","):
					print_img += "\t" 
				else:
					print_err("インタプリタエラー:print文")
					return -1
				tkn = tokens.pop(0)
				state = 0
				continue
		
		print(print_img)
		sendMsg(print_img)
		return 0

	
	if tkn.type == Tk.VAR: #代入文
		var = tkn
		dim = []
		tkn = tokens.pop(0)
		if not tkn.isRes("("):
			if var.image in var_dim.keys():
				print_err("K001:配列の添字指定がない: "+var.image)
				return -1
		else:
			#配列の代入文
			#  代入文：   変数 '(' 添字 [, 添字] ')' = 算術式
			#  添字：     { 数字変数 | 数字定数 }
			if not var.image in var_dim.keys():
				print_err("K002:配列として定義されていない: "+var.image)
				return -1
			tkn = tokens.pop(0)
			state = 0
			while True:
				if tkn.type == Tk.END:
					print_err("K003:配列の添字指定が正しくない: "+tkn.image)
					return -1
				if state == 0:
					tokens.insert(0, tkn)
					ret, tkns = copy_until(tokens,[",",")"])
					if ret == -1:
						print_err("K011:「)」または「,」でなければならない: "+tkn.image)
						return -1
					val = expression(tkns)
					if val == None:
						return -1
					dim.append(int(val))
					if len(dim) > 2:
						print_err("K006:3次元以上は指定できない: "+var.image)
						return -1
					state = 1
					tkn = tokens.pop(0)
					continue
				if state == 1:
					if tkn.isRes(","):
						tkn = tokens.pop(0)
						state = 0
						continue
					if tkn.isRes(")"):
						if len(dim) != len(var_dim[var.image]):
							print_err("K009:定義されている次元数("+str(len(var_dim[var.image]))+")が指定されていない: "+var.image)
							return -1
						for i, j in zip(var_dim[var.image], dim):
							if i < j or j <= 0:
								print_err("K010:添字の値が1から最大値("+str(int(i))+")の範囲ではない: "+str(int(j)))
								return -1
						tkn = tokens.pop(0)
						break
					print_err("K011:「)」または「,」でなければならない: "+tkn.image)
					return -1
		if tkn.isRes("="):
			if var.image[-1] != "$":
				ans =  expression(tokens)
			else:
				ans = expression_lit(tokens)
			if ans != None:
				if len(dim) == 0:
					assign[var.image] = ans
				elif len(dim) == 1:
					assign[var.image][dim[0]-1] = ans
				else:
					assign[var.image][dim[0]-1][dim[1]-1] = ans
				return 0
			else:
				return -1
		else:
			print_err("K012: 代入文の「=」がない: "+tkn.image)
			return -1

	print_err("Z001: サポートされていない命令である: "+tkn.image)
	return -1
	
def renumber(st):
	'''
	renum [ 開始行番号 ] [ , 増分 ]
	      0              1   2
	'''
	global prg
	
	#renumコマンド解析
	tokens = make_token(st)
	if tokens == None:
		print("Usage: renum [新行番号][,増分]")
		return prg
	tkn = tokens.pop(0)
	if tkn.type != Tk.VAR or tkn.image.upper() != "RENUM":
		print("Usage: renum [新行番号][,増分]")
		return prg
	state = 0
	tkn = tokens.pop(0)
	start_line = None
	inc_line = None
	while tkn.type != Tk.END:
		if state == 0:
			if tkn.type == Tk.NUMLIT:
				if "." in tkn.image:
					print("新行番号は整数でなければならない")
					return prg
				start_line = int(tkn.image)
				state = 1
				tkn = tokens.pop(0)
				continue
			if tkn.isRes(","):
				start_line = 10
				state = 2
				tkn = tokens.pop(0)
				continue
			print("Usage: renum [新行番号][,増分]")
			return prg
		if state == 1:
			if tkn.isRes(","):
				tkn = tokens.pop(0)
				state = 2
				continue
			print("Usage: renum [新行番号][,増分]")
			return prg
		if state == 2:
			if tkn.type == Tk.NUMLIT:
				if "." in tkn.image:
					print("増分は整数でなければならない")
					return prg
				inc_line = int(tkn.image)
				break
			print("Usage: renum [新行番号][,増分]")
			return prg
	if start_line == None:
		start_line = 10
	if inc_line == None:
		inc_line = 10
		
	#新行番号付け
	if prg.set_first() == -1:
		print("プログラムがない")
		return prg
	new_prg = Program()
	new_line = {} # 旧行番号と新行番号の対比表  new_line[旧行番号] = 新行番号					
	while True:
		line, tokens = prg.get_line()
		new_prg.put(start_line, tokens)
		new_line[line] = start_line
		start_line += inc_line
		if prg.set_next() == -1:
			break
	new_prg.set_first()
	while True:
		line, tokens = new_prg.get_line()
		tkn = tokens.pop(0)
		if not tkn.isRes("goto") and not tkn.isRes("gosub") and\
			not tkn.isRes("if"):
			if new_prg.set_next() == -1:
				break
			continue
		new_tokens = [tkn]
		if tkn.isRes("goto") or tkn.isRes("gosub"):
			pass
		elif tkn.isRes("if"):
			tkn = tokens.pop(0)
			while tkn.type != Tk.END and not tkn.isRes("goto"):
				new_tokens.append(tkn)
				tkn = tokens.pop(0)
			new_tokens.append(tkn)
			if tkn.type == Tk.END:
				if new_prg.set_next() == -1:
					break
				continue
		tkn = tokens.pop(0)
		if tkn.type == Tk.NUMLIT:
			if int(float(tkn.image)) in new_line.keys():
				tkn.image = str(new_line[int(float(tkn.image))])
		new_tokens.append(tkn)
		new_tokens.extend(tokens)
		new_prg.put(line, new_tokens)
		if new_prg.set_next() == -1:
			break
		continue
	return new_prg
	
def nest_check(curr_nest, tokens):
	nest_add = 0
	tkn = tokens.pop(0)
	if tkn.isRes("if"):
		tkn = tokens.pop()
		tkn = tokens.pop()
		if tkn.isRes("then"):
			nest_add = 1
	elif tkn.isRes("endif") or tkn.isRes("next"):
		curr_nest -= 1
	elif tkn.isRes("else"):
		curr_nest -= 1
		nest_add = 1
	elif tkn.isRes("for"):
		nest_add = 1
	return curr_nest, nest_add
def sendMsg(msg):
        ser.write(msg.encode('utf-8'))
        time.sleep(0.5)
def main():
        global print_stk, assign, trace_flg, running_flg, prg, var_dim,ser
        print("Open Port")
        ser=serial.Serial("COM**",115200)  ### "COM**" は「uart-mz-basic.ino」のシリアルポート番号を設定する
        print_tkn = False
        print_stk = False
        trace_flg = False
        running_flg = False
        assign = {}
        var_dim = {}
        prg = Program()
	
        while True:
        	st = input(">")
        	sendMsg(">"+st)
        	if st == "":
        		continue
        	if st.upper() == "QUIT":
        		break
        	if st.upper() == "TOKEN ON":
        		print_tkn = True
        		continue
        	if st.upper() == "TOKEN OFF":
        		print_tkn = False
        		continue
        	if st.upper() == "STACK ON":
        		print_stk = True
        		continue
        	if st.upper() == "STACK OFF":
        		print_stk = False
        		continue
        	if st.upper() == "PRINT_VAR":
        		print("assign=",assign)
        		print("var_dim=",var_dim)
        		continue
        	if st.upper() == "TRACE ON":
        		trace_flg = True
        		continue
        	if st.upper() == "TRACE OFF":
        		trace_flg = False
        		continue
        	if st.upper() == "LIST":
        		curr_nest = 0
        		if prg.set_first() == -1:
        			continue
        		while True:
        			_, tokens = prg.get_line()
        			curr_nest, nest_add = nest_check(curr_nest, tokens)
        			print(prg.line_image(curr_nest))
        			sendMsg(prg.line_image(curr_nest))
        			curr_nest += nest_add
        			if prg.set_next() == -1:
        				break
        		continue
        	if "RENUM" in st.upper():
        		prg = renumber(st)
        		continue
        	if "LOAD" in st.upper():

        		load_file = st.split(' ')
        		if len(load_file) != 2:
        			print("Usage: load file-name")
        			continue
        		infile = load_file[1]
        		try:
        			with open(infile, "r") as f:
        				lines = f.readlines()
        		except Exception as e:
        			print("ロードできませんでした: ",e)
        			continue
        		prg.clear()
        		for line in lines:
        			if line == "":
        				continue
        			tokens = make_token(line)
        			if tokens == None:
        				continue
        			tkn = tokens.pop(0)
        			if tkn.type == Tk.NUMLIT:
        				prg.put(int(float(tkn.image)), tokens)
        		continue	
        	if "SAVE" in st.upper():
        		save_file = st.split(' ')
        		if len(save_file) != 2:
        			print("Usage: save file-name")
        			continue
        		outfile = save_file[1]
        		try:
        			with open(outfile, "w") as f:
        				if prg.set_first() == -1:
        					print("プログラムがありません")
        					continue
        				curr_nest = 0
        				while True:
        					_, tokens = prg.get_line()
        					curr_nest, nest_add = nest_check(curr_nest, tokens)
        					f.write(prg.line_image(curr_nest)+"\n")
        					curr_nest += nest_add
        					if prg.set_next() == -1:
        						break
        		except Exception as e:
        			print("セーブできませんでした: ",e)
        			continue
        		continue
        	if st.upper() == "RUN":
        		if prg.set_first() == -1:
        			print("実行するプログラムがありません")
        			continue
        		running_flg = True
        		while True:
        			_, tokens = prg.get_line()
        			ret = statement(tokens)
        			if ret == 0: #次の行の実行
        				if prg.set_next() == -1:
        					break
        				continue
        			if ret == -1: #終了（STOP文とか）
        				break
        			#retは次に実行する行番号の場合
        			if prg.set_line(ret) == -1:
        				print_err("R001:この行番号は存在しない: "+str(ret))
        				break
        		running_flg = False
        		continue
        	#トークン列を作成する
        	tokens = make_token(st)
        	if tokens == None:
        		continue
        	if print_tkn:
        		for token in tokens:
        			print(token.type, token.image)
        	#トークン列を使って計算する。
        	tkn = tokens.pop(0)
        	if tkn.type == Tk.NUMLIT: #行番号
        		if "." in tkn.image:
        			print("行番号が正しくない")
        			continue
        		if len(tokens) == 1: # Tk.ENDのときのみ
        			prg.del_line(int(tkn.image))
        			continue
        		prg.put(int(tkn.image), tokens)
        		continue
        	if tkn.isRes("if") or tkn.isRes("for") or tkn.isRes("goto") or\
        		tkn.isRes("gosub"):
        		print("制御文は直接実行できません")
        		continue
        	tokens.insert(0, tkn)
        	statement(tokens)

if __name__ == "__main__":
	main()
